import Scope from './Scope'
import fetch from './fetch'
import debug from '../debug'
import * as functions from './functions'
import * as types from './types'
import sort from './sort'
import match from './match'

const DEFAULT_LIMIT = 100

// Implementation of equals that do not respect object equality and eschews deep compare
function equals(lhs, rhs) {
  if (typeof lhs === 'object' || typeof rhs === 'object') {
    if (lhs && rhs && lhs.toJSON && rhs.toJSON) {
      return lhs.toJSON() == rhs.toJSON()
    }
    return false
  }
  return lhs == rhs
}

function lt(lhs, rhs) {
  if (typeof lhs === 'object' || typeof rhs === 'object') {
    if (lhs && rhs && lhs.toJSON && rhs.toJSON) {
      return lhs.toJSON() < rhs.toJSON()
    }
    return false
  }
  return lhs < rhs
}

function gt(lhs, rhs) {
  if (typeof lhs === 'object' || typeof rhs === 'object') {
    if (lhs && rhs && lhs.toJSON && rhs.toJSON) {
      return lhs.toJSON() > rhs.toJSON()
    }
    return false
  }
  return lhs > rhs
}

module.exports = class Executor {

  constructor(options) {
    const {operations, scope, fetcher} = options
    this.scope = scope || new Scope({
      parent: null,
      value: {}
    })
    this.fetcher = fetcher
    this.operations = operations
  }

  async run() {
    return this.exec(this.operations, this.scope)
  }

  async exec(operation, scope) {
    debug('exec() op:', operation, 'scope:', scope)
    if (scope instanceof Promise) {
      throw new Error('scopes should be plain values, not promises')
    }
    switch (operation.op) {
      case 'pipe':
        return this.execPipeline(operation, scope)
      case 'source':
        const data = scope.dataForSource(operation.id)
        return scope.child({
          sourceId: operation.id,
          value: data.documents,
          start: data.start
        })
      case 'filter':
        return this.execFilter(operation, scope)
      case 'and':
        return this.execAnd(operation, scope)
      case 'or':
        return this.execOr(operation, scope)
      case 'eq':
        return this.execEq(operation, scope)
      case 'neq':
        return this.execNEq(operation, scope)
      case 'lt':
        return this.execLT(operation, scope)
      case 'lte':
        return this.execLTE(operation, scope)
      case 'gt':
        return this.execGT(operation, scope)
      case 'gte':
        return this.execGTE(operation, scope)
      case 'not':
        return this.execNOT(operation, scope)
      case 'accessor':
        return this.execAccessor(operation, scope)
      case 'literal':
        return this.execLiteral(operation, scope)
      case 'object':
        return this.execObjectExpr(operation, scope)
      case 'array':
        return this.execArray(operation, scope)
      case 'in':
        return this.execIn(operation, scope)
      case 'functionCall':
        return this.execFunctionCall(operation, scope)
      case 'match':
        return this.execMatch(operation, scope)
      default:
        throw new Error(`(exec) Unknown operation ${operation.op}`)
    }
  }

  async execFunctionCall(operation, scope) {
    const fn = functions[operation.name]
    if (!fn) {
      throw new Error(`Unknown function ${operation.name}`)
    }
    const scopedArgs = await Promise.all(
      operation.arguments.map(arg => this.exec(arg, scope))
    )
    return scope.clone({
      value: fn(scope, ...scopedArgs)
    // value: fn(...(scopedArgs.map(scopedArg => scopedArg.value)))
    })
  }

  // Exec an array of operations and return their plain values as an array
  async execEval(operations, scope) {
    return Promise.all(operations.map(op => this.exec(op, scope))).then(scopes => {
      return scopes.map(s => s.value)
    })
  }

  async execMatch(operation, scope) {
    const [lhs, rhs] = await Promise.all([
      this.exec(operation.lhs, scope),
      this.exec(operation.rhs, scope)
    ])
    const terms = Array.isArray(rhs.value) ? rhs.value : [rhs.value]
    return scope.clone({
      value: match(lhs.value, terms)
    })

  }

  async execIn(operation, scope) {
    const [lhs, rhs] = await Promise.all([
      this.exec(operation.lhs, scope),
      this.exec(operation.rhs, scope)
    ])
    if (lhs.value === null || lhs.value === undefined || rhs.value === null || rhs.value === undefined) {
      return scope.clone({
        value: null
      })
    }
    if (Array.isArray(rhs.value)) {
      for (let i = 0; i < rhs.value.length; i++) {
        const candidate = rhs.value[i]
        if (lhs.value == candidate) {
          return scope.clone({
            value: true
          })
        }
        if (candidate instanceof types.Path) {
          if (candidate.contains(lhs.value)) {
            return scope.clone({
              value: true
            })
          }
        }
      }
      return scope.clone({
        value: false
      })

    }
    if (rhs.value instanceof types.Path) {
      return scope.clone({
        value: rhs.value.contains(lhs.value)
      })
    }
    throw new Error(`in-operator does not apply to rhs value ${rhs.value}`)
  }


  async execArray(operation, scope) {
    return Promise.all(operation.operations.map(op => this.exec(op, scope))).then(elements => {
      return scope.clone({
        value: elements.map(element => element.value)
      })
    })
  }

  async execObjectExpr(operation, scope) {
    const evalObjOp = async (objOp) => {
      debug('objOp:', objOp)
      switch (objOp.op) {
        case 'assignment':
          return this.exec(objOp.value, scope).then(resultScope => {
            return {
              [objOp.name]: resultScope.value
            }
          })
        case 'splat':
          return scope.value
        default:
          throw new Error(`Unknown object expression operation ${objOp.op}`)
      }
    }

    return Promise.all(operation.operations.map(evalObjOp)).then(fragments => {
      debug('object fagments:', fragments)
      const value = Object.assign({}, ...fragments)
      const result = scope.clone({
        value
      })
      debug('projected object:', result)
      return result
    })
  }

  async pipeMap(operation, lhs) {
    let arrayValue = Array.isArray(lhs.value) ? lhs.value : [lhs.value]
    const result = arrayValue.map(item => {
      const scope = lhs.clone({
        value: item
      })
      return this.exec(operation, scope)
    })
    return Promise.all(result).then(result => {
      return lhs.clone({
        value: result.map(scope => scope.value)
      })
    })
  }

  async pipeFilter(operation, lhs) {
    let arrayValue = Array.isArray(lhs.value) ? lhs.value : [lhs.value]
    const filterPromises = arrayValue.map(item => {
      const scope = lhs.clone({
        value: item
      })
      return this.exec(operation, scope)
    })
    return Promise.all(filterPromises).then(
      filterValues => {
        debug('filter result:', filterValues)
        return lhs.clone({
          value: arrayValue.filter(((element, index) => filterValues[index].value === true))
        })
      }
    )
  }

  async pipe(operation, lhs) {
    const inputIsArray = Array.isArray(lhs.value)
    let arrayValue = inputIsArray ? lhs.value : [lhs.value]
    let result
    switch (operation.op) {
      case 'subscript':
        debug('subscript:', operation)
        result = lhs.slice(operation.start, operation.end)
        if (operation.first) {
          result = lhs.clone({
            value: result.value[0],
            start: null
          })
        }
        break
      case 'ordering':
        result = sort(lhs, operation.terms)
        break
      case 'filter':
        result = this.pipeFilter(operation, lhs)
        break
      default:
        debug('mapping:', lhs)
        result = this.pipeMap(operation, lhs).then(result => {
          debug('mapped:', result)
          return result
        })
        break
    }
    if (inputIsArray) {
      return result
    }
    if (!result.then) {
      return Array.isArray(result) ? result[0] : result
    }
    return result.then(arrayResult => {
      return arrayResult.clone({
        value: arrayResult.value[0]
      })
    })
  }

  // NOTE: Consider keeping the pipeline value as an array of scope, not a scope
  // containing an array. Have a feeling each value in the mapping chain needs
  // the ability to keep track of it's own parents etc.
  async execPipeline(operation, scope) {
    scope = await fetch(scope, operation, this.fetcher)
    const ops = operation.operations.slice()
    const first = ops.shift()
    let current = await this.exec(first, scope)
    while (ops.length > 0) {
      debug('current:', current)
      let nextOp = ops.shift()
      current = await this.pipe(nextOp, current)
    }
    debug('current:', current)
    return current
  }

  async execBinaryOp(operation, scope, impl) {
    const lhs = await this.exec(operation.lhs, scope)
    const rhs = await this.exec(operation.rhs, scope)
    return new Scope({
      parent: scope.parent,
      value: impl(lhs.value, rhs.value)
    })
  }

  async execPrefixOp(operation, scope, impl) {
    const rhs = await this.exec(operation.rhs, scope)
    return new Scope({
      parent: scope.parent,
      value: impl(rhs.value)
    })
  }

  async execAnd(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => {
      if (lhs === false || rhs === false) {
        return false
      }
      if (typeof lhs !== 'boolean' || typeof rhs !== 'boolean') {
        return null
      }
      return lhs && rhs
    })
  }

  async execOr(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => {
      if (lhs === true || rhs === true) {
        return true
      }
      if (typeof lhs !== 'boolean' || typeof rhs !== 'boolean') {
        return null
      }
      return lhs || rhs
    })
  }

  async execEq(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => equals(lhs, rhs))
  }

  async execNEq(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => !equals(lhs, rhs))
  }

  async execLT(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => lt(lhs, rhs))
  }

  async execLTE(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => lt(lhs, rhs) || equals(lhs, rhs))
  }

  async execGT(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => gt(lhs, rhs))
  }

  async execGTE(operation, scope) {
    return this.execBinaryOp(operation, scope, (lhs, rhs) => gt(lhs, rhs) || equals(lhs, rhs))
  }

  async execNOT(operation, scope) {
    return this.execPrefixOp(operation, scope, rhs => {
      if (typeof rhs !== 'boolean') {
        return null
      }
      return !rhs
    })
  }

  async execAccessor(operation, scope) {
    return scope.resolveAccessor(operation.path)
  }

  async execLiteral(operation, scope) {
    return new Scope({
      parent: null,
      value: operation.value
    })
  }

  async execFilter(operation, scope) {
    return this.exec(operation.filter, scope)
  }
}

